import polars as pl

from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()

with fs.transaction as tx:
    tx.create_branch("quickstart", "us-lakes", "main")

    lakes = pl.read_parquet("lakefs://quickstart/main/lakes.parquet")
    us_lakes = lakes.filter(pl.col("Country") == "United States of America")

    with fs.open("lakefs://quickstart/us-lakes/us_lakes.csv", "wb") as f:
        us_lakes.write_csv(f)

    tx.commit("quickstart", "us-lakes", "Add US lakes")



import duckdb

from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()
duckdb.register_filesystem(fs)

with fs.transaction as tx:
    tx.create_branch("quickstart", "italian-lakes", "main")

    lakes = duckdb.read_parquet("lakefs://quickstart/main/lakes.parquet")
    italian_lakes = duckdb.sql("SELECT * FROM lakes where Country='Italy'")
    italian_lakes.to_csv("lakefs://quickstart/italian-lakes/italian_lakes.csv")

    tx.commit("quickstart", "italian-lakes", "Add Italian lakes")



"""
This module provides a collection of functions to interact with a lakeFS server using the lakeFS SDK.
It includes functionalities to manage branches, repositories, commits, tags, and merge operations in
a lakeFS instance.

Note: Users of this module should have a configured and authenticated LakeFSClient instance, which is
required input to all functions.
"""

from __future__ import annotations

import json
import logging

from lakefs_sdk import BranchCreation
from lakefs_sdk.client import LakeFSClient
from lakefs_sdk.exceptions import ApiException, NotFoundException
from lakefs_sdk.models import (
    Commit,
    CommitCreation,
    Ref,
    Repository,
    RepositoryCreation,
    ResetCreation,
    RevertCreation,
    TagCreation,
)

from lakefs_spec.util import depaginate

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def commit(
    client: LakeFSClient,
    repository: str,
    branch: str,
    message: str,
    metadata: dict[str, str] | None = None,
) -> Commit:
    """
    Create a new commit of all uncommitted changes on the branch in the lakeFS file storage.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Repository name.
    branch: str
        Branch name.
    message: str
        Commit message.
    metadata: dict[str, str] | None
        Additional metadata for the commit.

    Returns
    -------
    Commit
        The created commit object of the lakeFS server.
    """
    diff = client.branches_api.diff_branch(repository=repository, branch=branch)

    if not diff.results:
        logger.warning(f"No changes to commit on branch {branch!r}, aborting commit.")
        return rev_parse(client, repository, branch, parent=0)

    commit_creation = CommitCreation(message=message, metadata=metadata)

    new_commit = client.commits_api.commit(
        repository=repository, branch=branch, commit_creation=commit_creation
    )
    return new_commit


def create_branch(
    client: LakeFSClient, repository: str, name: str, source_branch: str, exist_ok: bool = True
) -> str:
    """
    Create a branch ``name`` in a lakeFS repository.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Repository name.
    name: str
        Name of the branch to be created.
    source_branch: str
        Name of the source branch the new branch is created from.
    exist_ok: bool
        Ignore creation errors if the branch already exists.

    Returns
    -------
    str
        Name of newly created or existing branch.

    Raises
    ------
    ApiException
        If a branch with the same name already exists and ``exist_ok=False``.
    """

    try:
        new_branch = BranchCreation(name=name, source=source_branch)
        # client.branches_api.create_branch throws ApiException if branch exists
        client.branches_api.create_branch(repository=repository, branch_creation=new_branch)
        logger.debug(f"Created new branch {name!r} from branch {source_branch!r}.")
        return name
    except ApiException as e:
        if e.status == 409 and exist_ok:
            return name
        raise e


def delete_branch(
    client: LakeFSClient, repository: str, branch: str, missing_ok: bool = False
) -> None:
    """
    Delete the specified branch from a repository.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository from which the branch will be deleted.
    branch: str
        Name of the branch to be deleted.
    missing_ok: bool
        Ignore errors if the requested branch does not exist in the repository.

    Raises
    ------
    NotFoundException
        If the branch does not exist in the repository and ``missing_ok=False``.
    """
    try:
        client.branches_api.delete_branch(repository=repository, branch=branch)
    except NotFoundException as e:
        if not missing_ok:
            raise e


def list_branches(client: LakeFSClient, repository: str, prefix: str | None = None) -> list[Ref]:
    """
    List branches in a repository.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository for which to list branches.
    prefix: str | None
        Return branches prefixed with this value.

    Returns
    -------
    list[Ref]
        A list of qualified branches in the repository.
    """
    return list(depaginate(client.branches_api.list_branches, repository=repository, prefix=prefix))


def create_repository(
    client: LakeFSClient, name: str, storage_namespace: str, exist_ok: bool = True
) -> Repository:
    """
    Create a new repository in the lakeFS file storage system with a specified name and storage namespace.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    name: str
        Name of the repository to be created. This must be unique within the lakeFS instance.
    storage_namespace: str
        Storage namespace where the repository data will reside, typically corresponding to a bucket in object storage (e.g., S3 bucket) or a local namespace (e.g. local://<repo_name>).
    exist_ok: bool
        Ignore creation errors if the repository already exists.

    Returns
    -------
    Repository
        Repository object of the lakeFS SDK representing the newly created or existing repository.

    Raises
    ------
    ApiException
        If a repository of the same name already exists and ``exist_ok=False``.

    Notes
    -----
    Attempting to recreate a repository with the same name and storage namespace after previous deletion may lead to issues due to residual data, and is not recommended.
    """
    try:
        repository_creation = RepositoryCreation(name=name, storage_namespace=storage_namespace)
        return client.repositories_api.create_repository(repository_creation=repository_creation)
    except ApiException as e:
        if exist_ok:
            msg: str = json.loads(e.body)["message"]
            if e.status == 400 and msg.endswith("namespace already in use") or e.status == 409:
                return client.repositories_api.get_repository(name)
        raise e


def create_tag(client: LakeFSClient, repository: str, ref: str | Commit, tag: str) -> Ref:
    """
    Create a new tag in the specified repository in the lakeFS file storage system.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository where the tag will be created.
    ref: str | Commit
        Commit SHA or Commit object of the commit to which the tag will point.
    tag: str
        Name of the tag to be created.

    Raises
    ------
    ApiException
        If a tag of the same name already exists and points to a different ref.

    Returns
    -------
    Ref
        Ref object of the lakeFS SDK representing the tag.
    """

    if isinstance(ref, Commit):
        ref = ref.id
    tag_creation = TagCreation(id=tag, ref=ref)

    try:
        return client.tags_api.create_tag(repository=repository, tag_creation=tag_creation)
    except ApiException as e:
        if e.status == 409:
            target_commit = rev_parse(client=client, repository=repository, ref=ref)
            existing_tag = client.tags_api.get_tag(repository=repository, tag=tag)
            if existing_tag.commit_id == target_commit.id:
                return existing_tag
        raise e


def delete_tag(client: LakeFSClient, repository: str, tag: str | Ref) -> None:
    """
    Delete the specified tag from a repository.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository from which the tag will be deleted.
    tag: str | Ref
        Tag to be deleted.
    """
    if isinstance(tag, Ref):
        tag = tag.id
    client.tags_api.delete_tag(repository=repository, tag=tag)


def list_tags(client: LakeFSClient, repository: str) -> list[Ref]:
    """
    List all the tags in the specified repository in the lakeFS file storage system.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository from which to list the tags.

    Returns
    -------
    list[Ref]
        Ref objects of the tag in the repository.
    """
    return list(depaginate(client.tags_api.list_tags, repository))


def merge(client: LakeFSClient, repository: str, source_ref: str, target_branch: str) -> None:
    """
    Merges changes from a source reference to a target branch in a specified repository in the lakeFS file storage system.
    If no differences between source_ref and target_branch are found, the merge process is aborted.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository where the merge will occur.
    source_ref: str
        Source reference (branch name or ref/commit SHA) from which changes will be merged.
    target_branch: str
        Target branch to which changes will be merged.
    """
    diff = client.refs_api.diff_refs(
        repository=repository, left_ref=target_branch, right_ref=source_ref
    )
    if not diff.results:
        logger.warning("No difference between source and target. Aborting merge.")
        return
    client.refs_api.merge_into_branch(
        repository=repository, source_ref=source_ref, destination_branch=target_branch
    )


def revert(client: LakeFSClient, repository: str, branch: str, parent_number: int = 1) -> None:
    """
    Revert the commit on the specified branch to the parent specified by parent_number.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Repository in which the specified branch is located.
    branch: str
        Branch on which the commit should be reverted.
    parent_number: int, optional
        If there are multiple parents to a commit, specify to which parent the commit should be reverted.
        ``parent_number = 1`` (the default)  refers to the first parent commit of the current ``branch`` tip.
    """
    revert_creation = RevertCreation(ref=branch, parent_number=parent_number)
    client.branches_api.revert_branch(
        repository=repository, branch=branch, revert_creation=revert_creation
    )


def reset_branch(client: LakeFSClient, repository: str, branch: str) -> None:
    """
    Reset the specified branch to its head.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Repository in which the specified branch is located.
    branch: str
        Branch to reset.
    """
    reset_creation = ResetCreation(type="reset")
    client.branches_api.reset_branch(
        repository=repository, branch=branch, reset_creation=reset_creation
    )


def rev_parse(
    client: LakeFSClient,
    repository: str,
    ref: str | Commit,
    parent: int = 0,
) -> Commit:
    """
    Resolve a commit reference to the most recent commit or traverses the specified number of parent commits on a branch in a lakeFS repository.

    Parameters
    ----------
    client: LakeFSClient
        lakeFS client object.
    repository: str
        Name of the repository where the commit will be searched.
    ref: str | Commit
        Commit SHA or Commit object (with SHA stored in its ``id`` attribute) to resolve.
    parent: int
        Optionally parse a parent of ``ref`` instead of ``ref`` itself as indicated by the number.
        Must be non-negative. ``parent = 0`` (the default)  refers to ``ref`` itself.

    Returns
    -------
    Commit
        Commit object representing the resolved commit in the lakeFS repository.

    Raises
    ------
    ValueError
        - If ``parent`` is negative.
        - If the specified number of parent commits exceeds the actual number of available parents.
        - If the provided ``ref`` does not match any revision in the specified repository.
    """
    if parent < 0:
        raise ValueError("Parent number cannot be negative")
    try:
        if isinstance(ref, Commit):
            ref = ref.id
        revisions = list(
            depaginate(
                client.refs_api.log_commits, repository, ref, limit=True, amount=2 * (parent + 1)
            )
        )
        if len(revisions) <= parent:
            raise ValueError(
                f"unable to fetch revision {ref}~{parent}: "
                f"ref {ref!r} only has {len(revisions)} parents"
            )
        return revisions[parent]
    except NotFoundException:
        raise ValueError(f"{ref!r} does not match any revision in lakeFS repository {repository!r}")



from typing import Any

from lakefs_sdk.models import Commit

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory, with_counter


def test_transaction_commit(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    message = f"Add file {random_file.name}"

    with fs.transaction as tx:
        fs.put_file(lpath, rpath)
        assert len(tx.files) == 1
        # sha is a placeholder for the actual SHA created on transaction completion.
        sha = tx.commit(repository, temp_branch, message=message)
        # stack contains the file to upload, and the commit op.
        assert len(tx.files) == 2
        assert not sha.available

    assert sha.available

    commits = fs.client.refs_api.log_commits(
        repository=repository,
        ref=temp_branch,
    )
    latest_commit = commits.results[0]

    assert latest_commit.message == message
    assert latest_commit.id == sha.id


def test_transaction_tag(fs: LakeFSFileSystem, repository: str) -> None:
    try:
        # tag gets created on exit of the context.
        with fs.transaction as tx:
            sha = tx.rev_parse(repository, "main")
            tag = tx.tag(repository=repository, ref=sha, tag="v2")

        assert sha.available

        tags = fs.client.tags_api.list_tags(repository).results
        assert tags[0].id == tag
        assert tags[0].commit_id == sha.id
    finally:
        fs.client.tags_api.delete_tag(repository=repository, tag=tag)


def test_transaction_merge(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
    temporary_branch_context: Any,
) -> None:
    random_file = random_file_factory.make()

    with temporary_branch_context("transaction-merge-test") as new_branch:
        resource = f"{repository}/{new_branch}/{random_file.name}"
        message = "Commit new file"

        with fs.transaction as tx:
            # stage a file on new_branch...
            fs.put_file(str(random_file), resource)
            # ... commit it with the above message
            tx.commit(
                repository=repository,
                branch=new_branch,
                message=message,
            )
            # ... and merge it into temp_branch.
            tx.merge(repository=repository, source_ref=new_branch, into=temp_branch)

        # at last, verify temp_branch@HEAD is the merge commit.
        commits = fs.client.refs_api.log_commits(
            repository=repository,
            ref=temp_branch,
        )
        latest_commit = commits.results[0]
        assert latest_commit.message == f"Merge {new_branch!r} into {temp_branch!r}"
        second_latest_commit = commits.results[1]
        assert second_latest_commit.message == message


def test_transaction_revert(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    message = f"Add file {random_file.name}"

    with fs.transaction as tx:
        fs.put_file(lpath, rpath, autocommit=False)
        tx.commit(repository, temp_branch, message=message)
        tx.revert(repository=repository, branch=temp_branch)

    commits = fs.client.refs_api.log_commits(
        repository=repository,
        ref=temp_branch,
    )
    latest_commit = commits.results[0]
    assert latest_commit.message == f"Revert {temp_branch}"


def test_transaction_branch(fs: LakeFSFileSystem, repository: str) -> None:
    branch = "new-hello"

    try:
        with fs.transaction as tx:
            tx.create_branch(repository=repository, name=branch, source_branch="main")

        branches = [
            b.id for b in fs.client.branches_api.list_branches(repository=repository).results
        ]

        # existence check for a newly created branch.
        assert branch in branches
    finally:
        fs.client.branches_api.delete_branch(
            repository=repository,
            branch=branch,
        )


def test_transaction_entry(fs: LakeFSFileSystem) -> None:
    fs.start_transaction()
    assert fs._intrans
    assert fs._transaction is not None


def test_transaction_failure(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    message = f"Add file {random_file.name}"

    fs.client, counter = with_counter(fs.client)
    try:
        with fs.transaction as tx:
            fs.put_file(lpath, rpath)
            tx.commit(repository, temp_branch, message=message)
            raise RuntimeError("something went wrong")
    except RuntimeError:
        pass

    # assert that no commit happens because of the exception.
    assert counter.count("commits_api.commit") == 0


def test_placeholder_representations(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    message = f"Add file {random_file.name}"

    with fs.transaction as tx:
        fs.put_file(lpath, rpath)
        sha = tx.commit(repository, temp_branch, message=message)
    latest_commit = fs.client.refs_api.log_commits(repository=repository, ref=temp_branch).results[
        0
    ]
    assert isinstance(sha, Commit)
    assert sha.id == latest_commit.id
    assert repr(sha.id) == repr(latest_commit.id)



import contextlib
import logging
import random
import string
import sys
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Generator, TypeVar
from unittest.mock import MagicMock

import pytest
import yaml
from lakefs_sdk import Configuration
from lakefs_sdk.client import LakeFSClient
from lakefs_sdk.models import BranchCreation, RepositoryCreation

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory

_TEST_REPO = "lakefs-spec-tests"

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))

T = TypeVar("T")

YieldFixture = Generator[T, None, None]


@dataclass
class LakeFSOptions:
    host: str
    username: str
    password: str


@pytest.fixture(scope="session")
def lakefs_options() -> LakeFSOptions:
    """Raw configuration options for a lakeFS test instance."""
    return LakeFSOptions(
        host="localhost:8000",
        username="AKIAIOSFOLQUICKSTART",
        password="wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY",
    )


@pytest.fixture
def fs(lakefs_options: LakeFSOptions) -> LakeFSFileSystem:
    LakeFSFileSystem.clear_instance_cache()
    return LakeFSFileSystem(**asdict(lakefs_options))


@pytest.fixture(scope="session")
def lakefs_client(lakefs_options: LakeFSOptions) -> LakeFSClient:
    """A lakeFS client for API operations outside of the file system."""

    configuration = Configuration(
        host=lakefs_options.host,
        username=lakefs_options.username,
        password=lakefs_options.password,
    )
    return LakeFSClient(configuration=configuration)


@pytest.fixture(scope="session")
def ensurerepo(lakefs_client: LakeFSClient) -> str:
    # no loop, assumes there exist fewer than 100 repos.
    reponames = [r.id for r in lakefs_client.repositories_api.list_repositories().results]

    if _TEST_REPO in reponames:
        logger.info(f"Test repository {_TEST_REPO!r} already exists.")
    else:
        storage_config = lakefs_client.config_api.get_config().storage_config
        storage_namespace = f"{storage_config.default_namespace_prefix}/{_TEST_REPO}"
        logger.info(
            f"Creating test repository {_TEST_REPO!r} "
            f"with associated storage namespace {storage_namespace!r}."
        )
        lakefs_client.repositories_api.create_repository(
            RepositoryCreation(
                name=_TEST_REPO,
                storage_namespace=storage_namespace,
                sample_data=True,
            )
        )
    return _TEST_REPO


@pytest.fixture(scope="session")
def repository(ensurerepo: str) -> str:
    return ensurerepo


@pytest.fixture
def temporary_branch_context(lakefs_client: LakeFSClient, repository: str) -> Any:
    @contextlib.contextmanager
    def _wrapper(name: str) -> YieldFixture[str]:
        try:
            lakefs_client.branches_api.create_branch(
                repository=repository,
                branch_creation=BranchCreation(name=name, source="main"),
            )
            yield name
        finally:
            lakefs_client.branches_api.delete_branch(
                repository=repository,
                branch=name,
            )

    return _wrapper


@pytest.fixture
def temp_branch(lakefs_client: LakeFSClient, repository: str) -> YieldFixture[str]:
    """Create a temporary branch for a test."""
    name = "test-" + "".join(random.choices(string.digits, k=8))
    try:
        lakefs_client.branches_api.create_branch(
            repository=repository,
            branch_creation=BranchCreation(name=name, source="main"),
        )
        yield name
    finally:
        lakefs_client.branches_api.delete_branch(
            repository=repository,
            branch=name,
        )


@pytest.fixture
def random_file_factory(tmp_path: Path) -> RandomFileFactory:
    return RandomFileFactory(path=tmp_path)


@pytest.fixture
def temporary_lakectl_config() -> YieldFixture[str]:
    d = {
        "credentials": {"access_key_id": "hello", "secret_access_key": "world"},
        "server": {"endpoint_url": "http://hello-world-xyz"},
    }

    loc = "~/.lakectl.yaml"
    path = Path(loc).expanduser()
    backup_path = path.with_stem(path.stem + "_BAK")

    try:
        if path.exists():
            path.rename(backup_path)
        with open(path, "w") as f:
            yaml.dump(d, f)
        yield loc
    finally:
        path.unlink()
        if backup_path.exists():
            backup_path.rename(path)


@pytest.fixture()
def mock_urlopen(monkeypatch: pytest.MonkeyPatch) -> MagicMock:
    mock = MagicMock()
    monkeypatch.setattr("urllib.request.urlopen", mock)
    return mock



import random
import string
import uuid
from typing import Any

import pytest
from lakefs_sdk.exceptions import ApiException, NotFoundException
from lakefs_sdk.models import BranchCreation

import lakefs_spec.client_helpers as client_helpers
from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory, commit_random_file_on_branch, list_branch_names


def test_create_tag(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    commit_random_file_on_branch(
        random_file_factory=random_file_factory,
        fs=fs,
        repository=repository,
        temp_branch=temp_branch,
    )

    tagname = f"Change_{uuid.uuid4()}"
    try:
        tag = client_helpers.create_tag(
            client=fs.client, repository=repository, ref=temp_branch, tag=tagname
        )

        assert tag in client_helpers.list_tags(fs.client, repository)
        existing_tag = client_helpers.create_tag(
            client=fs.client, repository=repository, ref=temp_branch, tag=tagname
        )
        assert tag == existing_tag
    finally:
        fs.client.tags_api.delete_tag(repository=repository, tag=tagname)


def test_cannot_reassign_tag(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    commit_random_file_on_branch(
        random_file_factory=random_file_factory,
        fs=fs,
        repository=repository,
        temp_branch=temp_branch,
    )

    tagname = f"Change_{uuid.uuid4()}"
    try:
        client_helpers.create_tag(
            client=fs.client, repository=repository, ref=temp_branch, tag=tagname
        )
        commit_random_file_on_branch(
            random_file_factory=random_file_factory,
            fs=fs,
            repository=repository,
            temp_branch=temp_branch,
        )
        with pytest.raises(ApiException) as e:
            client_helpers.create_tag(
                client=fs.client, repository=repository, ref=temp_branch, tag=tagname
            )
            assert e.status == 409
    finally:
        client_helpers.delete_tag(client=fs.client, repository=repository, tag=tagname)


def test_delete_tag(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    commit_random_file_on_branch(
        random_file_factory=random_file_factory,
        fs=fs,
        repository=repository,
        temp_branch=temp_branch,
    )

    tagname = f"Change_{uuid.uuid4()}"
    tag = client_helpers.create_tag(
        client=fs.client, repository=repository, ref=temp_branch, tag=tagname
    )
    assert tag in client_helpers.list_tags(fs.client, repository)
    client_helpers.delete_tag(client=fs.client, repository=repository, tag=tagname)
    assert tag not in client_helpers.list_tags(fs.client, repository)


def test_merge_into_branch(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
    temporary_branch_context: Any,
) -> None:
    random_file = random_file_factory.make()

    with temporary_branch_context("merge-into-branch-test") as new_branch:
        source_ref = f"{repository}/{new_branch}/{random_file.name}"
        fs.put(str(random_file), source_ref)
        client_helpers.commit(
            client=fs.client,
            repository=repository,
            branch=new_branch,
            message="Commit new file",
        )

        assert (
            len(
                fs.client.refs_api.diff_refs(
                    repository=repository, left_ref=temp_branch, right_ref=new_branch
                ).results
            )
            > 0
        )
        client_helpers.merge(
            client=fs.client,
            repository=repository,
            source_ref=new_branch,
            target_branch=temp_branch,
        )
        assert (
            len(
                fs.client.refs_api.diff_refs(
                    repository=repository, left_ref=temp_branch, right_ref=new_branch
                ).results
            )
            == 0
        )


def test_merge_into_branch_aborts_on_no_diff(
    caplog: pytest.LogCaptureFixture, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    client_helpers.merge(
        client=fs.client, repository=repository, source_ref="main", target_branch=temp_branch
    )
    assert caplog.records[0].message == "No difference between source and target. Aborting merge."


def test_create_repository(fs: LakeFSFileSystem) -> None:
    name = "testrepo" + "".join(random.choices(string.digits, k=8))

    storage_config = fs.client.config_api.get_config().storage_config
    namespace = f"{storage_config.default_namespace_prefix}/{name}"

    repo = None
    try:
        repo = client_helpers.create_repository(fs.client, name=name, storage_namespace=namespace)
        new_repo = client_helpers.create_repository(
            fs.client, name=name, storage_namespace=namespace
        )
        # exist_ok means the same existing repo is returned.
        assert repo == new_repo
    finally:
        if repo is not None:
            fs.client.repositories_api.delete_repository(repo.id)


def test_revert(
    fs: LakeFSFileSystem, random_file_factory: RandomFileFactory, repository: str, temp_branch: str
) -> None:
    random_file = random_file_factory.make()
    source_ref = f"{repository}/{temp_branch}/{random_file.name}"

    fs.put(str(random_file), source_ref)
    client_helpers.commit(
        client=fs.client, repository=repository, branch=temp_branch, message="Commit new file"
    )

    assert (
        len(
            fs.client.refs_api.diff_refs(
                repository=repository, left_ref="main", right_ref=temp_branch
            ).results
        )
        > 0
    )
    client_helpers.revert(client=fs.client, repository=repository, branch=temp_branch)
    assert (
        len(
            fs.client.refs_api.diff_refs(
                repository=repository, left_ref="main", right_ref=temp_branch
            ).results
        )
        == 0
    )


def test_rev_parse(
    fs: LakeFSFileSystem, random_file_factory: RandomFileFactory, repository: str, temp_branch: str
) -> None:
    current_head_commit = fs.client.refs_api.log_commits(
        repository=repository, ref=temp_branch
    ).results[0]
    random_file = random_file_factory.make()
    fs.put(str(random_file), f"{repository}/{temp_branch}/{random_file.name}")
    client_helpers.commit(
        client=fs.client, repository=repository, branch=temp_branch, message="New Commit"
    )

    next_head_commit = fs.client.refs_api.log_commits(
        repository=repository, ref=temp_branch
    ).results[0]

    assert (
        client_helpers.rev_parse(client=fs.client, repository=repository, ref=temp_branch, parent=0)
        == next_head_commit
    )
    assert (
        client_helpers.rev_parse(client=fs.client, repository=repository, ref=temp_branch, parent=1)
        == current_head_commit
    )


def test_rev_parse_error_on_commit_not_found(fs: LakeFSFileSystem, repository: str) -> None:
    non_existent_ref = f"{uuid.uuid4()}"
    with pytest.raises(
        ValueError,
        match=f"{non_existent_ref!r} does not match any revision",
    ):
        client_helpers.rev_parse(
            client=fs.client, repository=repository, ref=non_existent_ref, parent=0
        )


def test_rev_parse_error_on_parent_does_not_exist(
    fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    n_commits = len(fs.client.refs_api.log_commits(repository=repository, ref=temp_branch).results)
    non_existent_parent = n_commits + 1
    with pytest.raises(
        ValueError,
        match=f"unable to fetch revision {temp_branch}~{non_existent_parent}",
    ):
        client_helpers.rev_parse(
            client=fs.client, repository=repository, ref=temp_branch, parent=non_existent_parent
        )


def test_delete_branch(fs: LakeFSFileSystem, repository: str) -> None:
    temp_branch_name = f"Branch_{uuid.uuid4()}"
    fs.client.branches_api.create_branch(
        repository=repository,
        branch_creation=BranchCreation(name=temp_branch_name, source="main"),
    )
    assert temp_branch_name in list_branch_names(fs, repository)
    client_helpers.delete_branch(fs.client, repository=repository, branch=temp_branch_name)
    assert temp_branch_name not in list_branch_names(fs, repository)


def test_delete_undefined_branch_error(fs: LakeFSFileSystem, repository: str) -> None:
    not_existing_branch_name = f"Branch_{uuid.uuid4()}"
    with pytest.raises(NotFoundException):
        client_helpers.delete_branch(
            fs.client, repository=repository, branch=not_existing_branch_name, missing_ok=False
        )


def test_reset_branch(
    fs: LakeFSFileSystem,
    repository: str,
    random_file_factory: RandomFileFactory,
    temp_branch: str,
) -> None:
    lpath = random_file_factory.make()
    lpath.write_text("Hello")

    rpath = f"lakefs://{repository}/{temp_branch}/{lpath.name}"

    fs.put(str(lpath), rpath)
    client_helpers.reset_branch(fs.client, repository, temp_branch)
    with pytest.raises(FileNotFoundError):
        fs.info(rpath)



import pytest

from lakefs_spec.config import LakectlConfig


def test_config_read_nonexistent_file():
    """
    Tests that a FileNotFoundError is raised when a lakectl config is read from
    a nonexistent file.
    """
    with pytest.raises(FileNotFoundError):
        LakectlConfig.read("aaaaaaaaaaaaaaaa.yaml")



import json
from urllib.error import HTTPError

from lakefs_sdk.exceptions import ApiException

from lakefs_spec.errors import translate_lakefs_error


def test_error_translation() -> None:
    rpath = "repo/ref/ohno.txt"

    # first case: lakeFS API error, 401 unauthorized
    e = ApiException(status=401, reason="unauthorized")
    e.body = json.dumps({"message": "unauthorized"})

    translated_err = translate_lakefs_error(e, rpath=rpath)
    assert isinstance(translated_err, PermissionError)
    assert f"unauthorized: {rpath!r}" in str(translated_err)

    # second case: urllib 403 unauthorized (blockstore put)
    e = HTTPError("any", 403, "forbidden", None, None)  # type: ignore
    translated_err = translate_lakefs_error(e, rpath=rpath)
    assert isinstance(translated_err, PermissionError)
    assert f"forbidden: {rpath!r}" in str(translated_err)

    # third case: lakeFS API error 420 (corresponds to partial IOError)
    e = ApiException(status=420, reason="too many requests")
    e.body = json.dumps({"message": "too many requests"})
    translated_err = translate_lakefs_error(e, rpath=rpath)
    assert isinstance(translated_err, OSError)
    assert f"too many requests: {rpath!r}" in str(translated_err)

    # fourth case: lakeFS API error 400 with a custom message.
    e = ApiException(status=400, reason="bad request")
    e.body = json.dumps({"message": "bad request"})
    message = "oh no!"
    translated_err = translate_lakefs_error(e, message=message)
    assert isinstance(translated_err, OSError)
    assert str(translated_err).endswith(message)



from pathlib import Path

import pytest

from lakefs_spec.spec import LakeFSFileSystem
from tests.util import RandomFileFactory


def test_readline(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    """Checks that `fs.open()` behaves like builtin `open` for `readline(s)` APIs."""
    lpath = Path("random_file.txt")
    try:
        lpath.write_text("Hello\nmy name is\nxyz")
        with open(lpath, "rb") as f:
            native_open_line = f.readline()
            f.seek(0)
            native_open_lines = f.readlines()
            f.seek(0)
            firstline = next(f)

        rpath = f"{repository}/{temp_branch}/{Path(lpath).name}"
        fs.put_file(lpath, rpath)

        with fs.open(rpath, "rb") as rf:
            # mode == "rb" means everything is bytes.
            assert rf.readline() == native_open_line
            rf.seek(0)
            assert rf.readlines() == native_open_lines
            rf.seek(0)
            # default char is linebreak (b'\n'),
            # so `readuntil()` emulates the line iterator in its default state.
            assert rf.readuntil() == firstline
    finally:
        lpath.unlink(missing_ok=True)


def test_info(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    rnd_file = random_file_factory.make()
    rpath = f"{repository}/{temp_branch}/{rnd_file.name}"
    fs.put(str(rnd_file), rpath)
    details = fs.open(rpath).info()
    expected_keys = ["checksum", "content-type", "mtime", "name", "size", "type"]
    for key in expected_keys:
        assert key in details.keys()

    # opening a nonexistent file in read mode should immediately result in a ``FileNotFoundError``.
    with pytest.raises(FileNotFoundError):
        fs.open(f"{repository}/main/hello.tar.gz", "rb")


def test_readuntil(fs: LakeFSFileSystem, repository: str, temp_branch: str) -> None:
    lpath = Path("tmp_file.txt")

    content_first_part = (
        "This is a test file\nthat contains an occurence until which we want to read:7"
    )
    content_second_part = "\nand some words after it."
    try:
        lpath.write_text(content_first_part + content_second_part)
        rpath = f"{repository}/{temp_branch}/tmp_file.txt"
        fs.put_file(lpath, rpath)
        with fs._open(rpath, "rb") as rf:
            remote_readuntil = rf.readuntil(b"7").decode("utf-8")
            assert content_first_part == remote_readuntil
            assert content_second_part not in remote_readuntil
    finally:
        lpath.unlink(missing_ok=True)



from typing import ContextManager

import pytest

from lakefs_spec.spec import parse


@pytest.mark.parametrize(
    "path,repo,ref,resource",
    [
        # case 1: well-formed path input
        ("my-repo/my-ref/resource.txt", "my-repo", "my-ref", "resource.txt"),
        # case 2: nested resource
        ("my-repo/my-ref/my/nested/resource.txt", "my-repo", "my-ref", "my/nested/resource.txt"),
        # case 3: top-level resource
        ("my-repo/my-ref/", "my-repo", "my-ref", ""),
        # case 4: single-character branch
        ("my-repo/a/resource.txt", "my-repo", "a", "resource.txt"),
        # case 5: well-formed path with leading lakefs:// scheme
        ("lakefs://my-repo/my-ref/resource.txt", "my-repo", "my-ref", "resource.txt"),
    ],
)
def test_path_parsing_success(path: str, repo: str, ref: str, resource: str) -> None:
    act_repo, act_ref, act_resource = parse(path)
    assert act_repo == repo
    assert act_ref == ref
    assert act_resource == resource


@pytest.mark.parametrize(
    "path,expected_exception",
    [
        # repo name illegally begins with hyphen
        ("-repo/my-ref/resource.txt", pytest.raises(ValueError, match="expected path .*")),
        # repo name contains an illegal uppercase letter
        ("Repo/my-ref/resource.txt", pytest.raises(ValueError, match="expected path .*")),
        # missing repo name
        ("my-ref/resource.txt", pytest.raises(ValueError, match="expected path .*")),
        # illegal branch name
        ("repo/my-ref$$$/resource.txt", pytest.raises(ValueError, match="expected path .*")),
    ],
)
def test_path_parsing_failure(path: str, expected_exception: ContextManager) -> None:
    with expected_exception:
        parse(path)



from lakefs_spec import LakeFSFileSystem


def test_rm(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    path = f"{repository}/{temp_branch}/README.md"

    fs.rm(path)
    assert not fs.exists(path)


def test_rm_with_postcommit(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    path = f"{repository}/{temp_branch}/README.md"
    msg = "Remove file README.md"

    with fs.transaction as tx:
        fs.rm(path)
        tx.commit(repository=repository, branch=temp_branch, message=msg)
    assert not fs.exists(path)

    commits = fs.client.refs_api.log_commits(
        repository=repository,
        ref=temp_branch,
    )
    latest_commit = commits.results[0]
    assert latest_commit.message == msg



import random
import string

import pytest

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory, with_counter


def test_no_change_postcommit(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    with fs.transaction as tx:
        fs.put(lpath, rpath, precheck=False, autocommit=False)
        tx.commit(repository=repository, branch=temp_branch, message=f"Add file {random_file.name}")

    commits = fs.client.refs_api.log_commits(
        repository=repository,
        ref=temp_branch,
    )
    latest_commit = commits.results[0]  # commit log is ordered branch-tip-first
    assert latest_commit.message == f"Add file {random_file.name}"

    # put the same file again, this time the diff is empty
    with fs.transaction as tx:
        fs.put(lpath, rpath, precheck=False, autocommit=False)
        tx.commit(repository=repository, branch=temp_branch, message=f"Add file {random_file.name}")

    # check that no other commit has happened.
    commits = fs.client.refs_api.log_commits(
        repository=repository,
        ref=temp_branch,
    )
    assert commits.results[0] == latest_commit


def test_implicit_branch_creation(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    random_file = random_file_factory.make()
    lpath = str(random_file)

    non_existing_branch = "non-existing-" + "".join(random.choices(string.digits, k=8))
    rpath = f"{repository}/{non_existing_branch}/{random_file.name}"
    try:
        fs.put(lpath, rpath)
        branches = [
            r.id for r in fs.client.branches_api.list_branches(repository=repository).results
        ]
        assert non_existing_branch in branches
    finally:
        fs.client.branches_api.delete_branch(
            repository=repository,
            branch=non_existing_branch,
        )

    fs.create_branch_ok = False
    another_non_existing_branch = "non-existing-" + "".join(random.choices(string.digits, k=8))
    rpath = f"{repository}/{another_non_existing_branch}/{random_file.name}"
    with pytest.raises(FileNotFoundError, match=f"not found: {rpath!r}"):
        fs.put(lpath, rpath)


def test_put_client_caching(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    """
    Tests that ``precheck=True`` prevents a second upload of an identical file by matching checksums.
    """
    fs.client, counter = with_counter(fs.client)

    random_file = random_file_factory.make()
    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"
    fs.put(lpath, rpath)
    assert counter.count("objects_api.upload_object") == 1
    assert fs.exists(rpath)

    # second put, should not happen.
    fs.put(lpath, rpath)
    assert counter.count("objects_api.upload_object") == 1



"""
Functionality for extended lakeFS transactions to conduct versioning operations between file uploads.
"""

from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

import wrapt
from fsspec.spec import AbstractBufferedFile
from fsspec.transaction import Transaction
from lakefs_sdk.client import LakeFSClient
from lakefs_sdk.models import Commit, Ref

from lakefs_spec.client_helpers import commit, create_branch, create_tag, merge, rev_parse, revert

T = TypeVar("T")

if TYPE_CHECKING:
    from lakefs_spec import LakeFSFileSystem

    VersioningOpTuple = tuple[Callable[[LakeFSFileSystem], None], Any]


@dataclass
class Placeholder(Generic[T], wrapt.ObjectProxy):
    """A generic placeholder for a value computed by the lakeFS server in a versioning operation during a transaction."""

    def __init__(self, wrapped: T | None = None):
        super().__init__(wrapped)

    @property
    def available(self) -> bool:
        """Whether the wrapped value is available, i.e. already computed."""
        return self.__wrapped__ is not None

    @property
    def value(self):
        return self.__wrapped__

    @value.setter
    def value(self, val: T) -> None:
        """Fill in the placeholder. Not meant to be called directly except in the completion of the transaction."""
        self.__wrapped__ = val


class LakeFSTransaction(Transaction):
    """
    A lakeFS transaction model capable of versioning operations in between file uploads.

    Parameters
    ----------
    fs: LakeFSFileSystem
        The lakeFS file system associated with the transaction.
    """

    def __init__(self, fs: "LakeFSFileSystem"):
        super().__init__(fs=fs)
        self.fs: "LakeFSFileSystem"
        self.files: deque[AbstractBufferedFile | VersioningOpTuple] = deque(self.files)

    def __enter__(self):
        self.fs._intrans = True
        return self

    def commit(
        self, repository: str, branch: str, message: str, metadata: dict[str, str] | None = None
    ) -> Placeholder[Commit]:
        """
        Create a commit on a branch in a repository with a commit message and attached metadata.

        Parameters
        ----------
        repository: str
            The repository to create the commit in.
        branch: str
            The name of the branch to commit on.
        message: str
            The commit message to attach to the newly created commit.
        metadata: dict[str, str] | None
            Optional metadata to enrich the created commit with (author, e-mail, etc.).

        Returns
        -------
        Placeholder[Commit]
            A placeholder for the commit created by the dispatched ``commit`` API call.
        """
        # bind all arguments to the client helper function, and then add it to the file-/callstack.
        op = partial(
            commit, repository=repository, branch=branch, message=message, metadata=metadata
        )
        p: Placeholder[Commit] = Placeholder()
        self.files.append((op, p))
        # return a placeholder for the commit.
        return p

    def complete(self, commit: bool = True) -> None:
        """
        Finish the transaction by unwinding the file/versioning op stack via

        1. Committing or discarding in case of a file, and
        2. Conducting versioning operations using the file system's client.

        No operations happen and all files are discarded if ``commit == False``,
        which is the case, e.g., if an exception happens in the context manager.

        Parameters
        ----------
        commit: bool
            Whether to conduct operations queued in the transaction.
        """
        while self.files:
            # fsspec base class calls `append` on the file, which means we
            # have to pop from the left to preserve order.
            f = self.files.popleft()
            if isinstance(f, AbstractBufferedFile):
                if commit:
                    f.commit()
                else:
                    f.discard()
            else:
                # client helper + return value case.
                op, retval = f
                if commit:
                    result = op(self.fs.client)
                    # if the transaction member returns a placeholder,
                    # fill it with the result of the client helper.
                    if isinstance(retval, Placeholder):
                        retval.value = result

        self.fs._intrans = False

    def create_branch(
        self, repository: str, name: str, source_branch: str, exist_ok: bool = True
    ) -> str:
        """
        Create a branch ``name`` in a repository, branching off ``source_branch``.

        Parameters
        ----------
        repository: str
            Repository name.
        name: str
            Name of the branch to be created.
        source_branch: str
            Name of the source branch that the new branch is created from.
        exist_ok: bool
            Ignore creation errors if the branch already exists.

        Returns
        -------
        str
            The requested branch name.
        """
        op = partial(
            create_branch,
            repository=repository,
            name=name,
            source_branch=source_branch,
            exist_ok=exist_ok,
        )
        self.files.append((op, name))
        return name

    def merge(self, repository: str, source_ref: str, into: str) -> None:
        """
        Merge a branch into another branch in a repository.

        Parameters
        ----------
        repository: str
            Name of the repository.
        source_ref: str
            Source reference containing the changes to merge. Can be a branch name or partial commit SHA.
        into: str
            Target branch into which the changes will be merged.
        """
        op = partial(merge, repository=repository, source_ref=source_ref, target_branch=into)
        self.files.append((op, None))
        return None

    def revert(self, repository: str, branch: str, parent_number: int = 1) -> None:
        """
        Revert a previous commit on a branch.

        Parameters
        ----------
        repository: str
            Name of the repository.
        branch: str
            Branch on which the commit should be reverted.
        parent_number: int
            If there are multiple parents to a commit, specify to which parent the commit should be reverted.
            ``parent_number = 1`` (the default)  refers to the first parent commit of the current ``branch`` tip.
        """

        op = partial(revert, repository=repository, branch=branch, parent_number=parent_number)
        self.files.append((op, None))
        return None

    def rev_parse(
        self, repository: str, ref: str | Placeholder[Commit], parent: int = 0
    ) -> Placeholder[Commit]:
        """
        Parse a given reference or any of its parents in a repository.

        Parameters
        ----------
        repository: str
            Name of the repository.
        ref: str | Placeholder[Commit]
            Commit SHA or commit placeholder object to resolve.
        parent: int
            Optionally parse a parent of ``ref`` instead of ``ref`` itself as indicated by the number.
            Must be non-negative. ``parent = 0`` (the default)  refers to ``ref`` itself.

        Returns
        -------
        Placeholder[Commit]
            A placeholder for the commit created by the dispatched ``rev_parse`` API call.
        """

        def rev_parse_op(client: LakeFSClient, **kwargs: Any) -> Commit:
            return rev_parse(client, **kwargs)

        p: Placeholder[Commit] = Placeholder()
        op = partial(rev_parse_op, repository=repository, ref=ref, parent=parent)
        self.files.append((op, p))
        return p

    def tag(self, repository: str, ref: str | Placeholder[Commit], tag: str) -> str:
        """
        Create a tag referencing a commit in a repository.

        Parameters
        ----------
        repository: str
            Name of the repository.
        ref: str | Placeholder[Commit]
            Commit SHA or placeholder for a commit object to which the new tag will point.
        tag: str
            Name of the tag to be created.

        Returns
        -------
        str
            The name of the requested tag.
        """

        def tag_op(client: LakeFSClient, **kwargs: Any) -> Ref:
            return create_tag(client, **kwargs)

        self.files.append((partial(tag_op, repository=repository, ref=ref, tag=tag), tag))
        return tag



import pytest

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory, with_counter


@pytest.mark.parametrize("pagesize", [1, 2, 5, 10, 50])
def test_paginated_ls(fs: LakeFSFileSystem, repository: str, pagesize: int) -> None:
    """
    Check that all results of an ``ls`` call are returned independently of page size.
    """
    resource = f"{repository}/main/"

    # default amount of 100 objects per page
    all_results = fs.ls(resource)

    paged_results = fs.ls(resource, amount=pagesize, refresh=True)
    assert paged_results == all_results


def test_ls_caching(fs: LakeFSFileSystem, repository: str) -> None:
    """
    Check that ls calls are properly cached.
    """
    fs.client, counter = with_counter(fs.client)

    testdir = "data"
    resource = f"{repository}/main/{testdir}/"

    for _ in range(2):
        fs.ls(resource)
        assert len(fs.dircache) == 1
        assert set(fs.dircache.keys()) == {resource.removesuffix("/")}

    # assert the second `ls` call hits the cache
    assert counter.count("objects_api.list_objects") == 1


def test_ls_cache_refresh(fs: LakeFSFileSystem, repository: str) -> None:
    """
    Check that ls calls bypass the dircache if requested through ``refresh=False``
    """
    fs.client, counter = with_counter(fs.client)

    testdir = "data"
    resource = f"{repository}/main/{testdir}/"

    for _ in range(2):
        fs.ls(resource, refresh=True)
        assert len(fs.dircache) == 1
        assert set(fs.dircache.keys()) == {resource.removesuffix("/")}

    # assert the second `ls` call bypasses the cache
    assert counter.count("objects_api.list_objects") == 2


def test_ls_stale_cache_entry(
    fs: LakeFSFileSystem,
    repository: str,
    random_file_factory: RandomFileFactory,
    temp_branch: str,
) -> None:
    fs.client, counter = with_counter(fs.client)

    random_file = random_file_factory.make()

    resource = f"{repository}/main/data/"

    res = fs.ls(resource)
    assert counter.count("objects_api.list_objects") == 1
    assert set(fs.dircache.keys()) == {resource.removesuffix("/")}

    cache_entry = fs.dircache[resource.removesuffix("/")]
    assert len(cache_entry) == 1
    assert cache_entry[0] == res[0]

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/data/{random_file.name}"

    fs.put_file(lpath, rpath, precheck=False)

    res = fs.ls(rpath)
    assert counter.count("objects_api.list_objects") == 2

    # Should not be added to the cache for the other branch...
    assert len(cache_entry) == 1

    # ... but instead to the cache entry for the new branch
    cache_entry = fs.dircache[f"{repository}/{temp_branch}/data"]
    assert len(cache_entry) == 1
    assert cache_entry[0] == res[0]


def test_ls_no_detail(fs: LakeFSFileSystem, repository: str) -> None:
    fs.client, counter = with_counter(fs.client)

    branch = "main"
    prefix = f"{repository}/{branch}"
    resource = f"{prefix}/data"

    expected = [f"{resource}/lakes.source.md"]
    # first, verify the API fetch does the expected...
    assert fs.ls(resource, detail=False) == expected
    assert list(fs.dircache.keys()) == [resource]

    # ...as well as the cache fetch.
    assert fs.ls(resource, detail=False) == expected

    # One API call for the directory object, and one for listing its contents
    assert counter.count("objects_api.list_objects") == 2

    # test the same thing with a subfolder + file prefix
    resource = f"{prefix}/images/duckdb"
    fs.ls(resource, detail=False)

    assert set(fs.dircache.keys()) == {f"{prefix}/data", f"{prefix}/images"}


def test_ls_dircache_remove_uncached(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    prefix = f"{repository}/{temp_branch}"
    resource = f"{prefix}/"

    listing_pre = fs.ls(resource)
    fs.rm(listing_pre[0]["name"])

    # List again, bypassing the cache...
    listing_post = fs.ls(resource, refresh=True)
    assert len(listing_post) == len(listing_pre) - 1

    # ... and through the cache (which should have been updated above)
    listing_post = fs.ls(resource)
    assert len(listing_post) == len(listing_pre) - 1


def test_ls_dircache_remove_cached(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    prefix = f"{repository}/{temp_branch}"
    resource = f"{prefix}/"

    listing_pre = fs.ls(resource)
    fs.rm(listing_pre[0]["name"])

    # List again, cache should have been invalidated by rm
    listing_post = fs.ls(resource)
    assert len(listing_post) == len(listing_pre) - 1


def test_ls_dircache_recursive(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    prefix = f"{repository}/{temp_branch}"

    # (1) Basic recursive dircache filling
    listing = fs.ls(prefix + "/", recursive=True)
    assert len(fs.dircache) > 1  # Should contain entries for all sub-folders

    # Dircache invariant: all files in an entry must be direct descendants of its parent
    for cache_dir, files in fs.dircache.items():
        assert all([fs._parent(v["name"].rstrip("/")) == cache_dir for v in files])

    # (2) Dircache correctness, recursive
    cached_listing_recursive = fs.ls(prefix + "/", recursive=True)
    # Recursive listing from cache must contain all items, ordering need not be preserved
    assert {o["name"] for o in cached_listing_recursive} == {o["name"] for o in listing}

    # (3) Dircache correctness, non-recursive
    cached_listing_nonrecursive = fs.ls(prefix + "/", recursive=False)
    # Non-recursive listing from cache must only contain direct descendants of the listed directory
    # (and the subfolders directly contained therein)
    assert all([fs._parent(o["name"].rstrip("/")) == prefix for o in cached_listing_nonrecursive])

    # (4) Adding a file should only modify a single dircache entry
    directory = "data"
    filename = "new-file.txt"
    rpath = f"{prefix}/{directory}/{filename}"

    old_cache_len = len(fs.dircache.get(f"{prefix}/{directory}"))

    fs.pipe(rpath, b"data")
    _ = fs.ls(prefix + "/", refresh=True, recursive=True)

    cache_entry = fs.dircache.get(f"{prefix}/{directory}")

    # Added file appears in the cache entry for its parent dir
    assert len(cache_entry) == old_cache_len + 1
    assert rpath in {f["name"] for f in cache_entry}

    # Dircache invariant is maintained
    for cache_dir, files in fs.dircache.items():
        assert all([fs._parent(v["name"].rstrip("/")) == cache_dir for v in files])


def test_ls_directories(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    """Validate that recursive and non-recursive ``ls`` handle directory entries identically."""
    prefix = f"lakefs://{repository}/{temp_branch}"

    fs.rm(f"{prefix}/images/", recursive=True)
    fs.rm(f"{prefix}/data/", recursive=True)

    fs.pipe(f"{prefix}/a.txt", b"a")
    fs.pipe(f"{prefix}/dir1/b.txt", b"b")
    fs.pipe(f"{prefix}/dir1/dir2/c.txt", b"c")

    # (1) - recursive ls includes virtual directory entries for all levels except the root
    ls_recursive = fs.ls(prefix + "/", recursive=True)

    dirs = [o for o in ls_recursive if o["type"] == "directory"]
    assert len(dirs) == 2  # includes `dir1/` and `dir1/dir2`, but not the root.

    # (2) - non-recursive ls only includes virtual dir entries on the same level
    ls_nonrecursive = fs.ls(prefix + "/", recursive=False)

    dirs = [o for o in ls_nonrecursive if o["type"] == "directory"]
    assert len(dirs) == 1  # only `dir1/`



"""lakefs-spec is an fsspec file system integration for the lakeFS data lake."""

from importlib.metadata import PackageNotFoundError, version

from .spec import LakeFSFile, LakeFSFileSystem
from .transaction import LakeFSTransaction

try:
    __version__ = version("lakefs-spec")
except PackageNotFoundError:
    # package is not installed
    pass



import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq

from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()

with fs.transaction as tx:
    tx.create_branch("quickstart", "partitioned-data", "main")

    lakes_table = pq.read_table("quickstart/main/lakes.parquet", filesystem=fs)

    ds.write_dataset(
        lakes_table,
        "quickstart/partitioned-data/lakes",
        filesystem=fs,
        format="csv",
        partitioning=ds.partitioning(pa.schema([lakes_table.schema.field("Country")])),
    )

    tx.commit("quickstart", "partitioned-data", "Add partitioned lakes data set")



import pytest

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory


def test_lakefs_file_open_read(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
    random_file_factory: RandomFileFactory,
) -> None:
    random_file = random_file_factory.make()
    with open(random_file, "rb") as f:
        orig_text = f.read()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"
    fs.put_file(lpath, rpath)

    # try opening the remote file
    with fs.open(rpath) as fp:
        text = fp.read()

    assert text == orig_text


def test_lakefs_file_open_write(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
    random_file_factory: RandomFileFactory,
) -> None:
    random_file = random_file_factory.make()
    with open(random_file, "rb") as f:
        orig_text = f.read()

    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    # try opening the remote file and writing to it
    with fs.open(rpath, "wb") as fp:
        fp.write(orig_text)

    # pulling the written file down again, using ONLY built-in open (!)
    lpath = random_file.with_name(random_file.name + "_copy")

    blocksize = fs.blocksize
    fs.blocksize = 256
    fs.get(rpath, str(lpath))
    fs.blocksize = blocksize

    with open(lpath, "rb") as f:
        new_text = f.read()

    # round-trip assert.
    assert new_text == orig_text


def test_lakefs_file_unknown_mode(fs: LakeFSFileSystem) -> None:
    """Test that a NotImplementedError is raised on unknown mode encounter."""

    with pytest.raises(NotImplementedError, match="unsupported mode .*"):
        fs.open("hello.py", mode="ab")



"""
Core interface definitions for file system interaction with lakeFS from Python,
namely the ``LakeFSFileSystem`` and ``LakeFSFile`` classes.
"""

from __future__ import annotations

import errno
import io
import logging
import mimetypes
import operator
import os
import urllib.error
import urllib.request
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import Any, Generator, Iterable, Literal, cast, overload

import fsspec.callbacks
from fsspec import filesystem
from fsspec.callbacks import _DEFAULT_CALLBACK
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem
from fsspec.utils import stringify_path
from lakefs_sdk import Configuration
from lakefs_sdk.client import LakeFSClient
from lakefs_sdk.exceptions import ApiException, NotFoundException
from lakefs_sdk.models import ObjectCopyCreation, ObjectStats, StagingMetadata

from lakefs_spec.client_helpers import create_branch
from lakefs_spec.config import LakectlConfig
from lakefs_spec.errors import translate_lakefs_error
from lakefs_spec.transaction import LakeFSTransaction
from lakefs_spec.util import depaginate, md5_checksum, parse

logger = logging.getLogger(__name__)


class LakeFSFileSystem(AbstractFileSystem):
    """
    lakeFS file system implementation.

    Instances of this class are cached based on their constructor arguments.

    For more information, see the fsspec documentation <https://filesystem-spec.readthedocs.io/en/latest/features.html#instance-caching>.

    Parameters
    ----------
    host: str | None
        The address of your lakeFS instance.
    username: str | None
        The access key name to use in case of access key authentication.
    password: str | None
        The access key secret to use in case of access key authentication.
    api_key: str | None
        The API key to use in case of authentication with an API key.
    api_key_prefix: str | None
        A string prefix to use for the API key in authentication.
    access_token: str | None
        An access token to use in case of access token authentication.
    verify_ssl: bool
        Whether to verify SSL certificates in API interactions. Do not disable in production.
    ssl_ca_cert: str | None
        A custom certificate PEM file to use to verify the peer in SSL connections.
    proxy: str | None
        Proxy address to use when connecting to a lakeFS server.
    configfile: str
        ``lakectl`` YAML configuration file to read credentials from.
    create_branch_ok: bool
        Whether to create branches implicitly when not-existing branches are referenced on file uploads.
    source_branch: str
        Source branch set as origin when a new branch is implicitly created.
    **storage_options: Any
        Configuration options to pass to the file system's directory cache.
    """

    protocol = "lakefs"

    def __init__(
        self,
        host: str | None = None,
        username: str | None = None,
        password: str | None = None,
        api_key: str | None = None,
        api_key_prefix: str | None = None,
        access_token: str | None = None,
        verify_ssl: bool = True,
        ssl_ca_cert: str | None = None,
        proxy: str | None = None,
        configfile: str = "~/.lakectl.yaml",
        create_branch_ok: bool = True,
        source_branch: str = "main",
        **storage_options: Any,
    ):
        super().__init__(**storage_options)

        if (p := Path(configfile).expanduser()).exists():
            lakectl_config = LakectlConfig.read(p)
        else:
            # empty config.
            lakectl_config = LakectlConfig()

        configuration = Configuration(
            host=host or os.getenv("LAKEFS_HOST") or lakectl_config.host,
            api_key=api_key or os.getenv("LAKEFS_API_KEY"),
            api_key_prefix=api_key_prefix or os.getenv("LAKEFS_API_KEY_PREFIX"),
            access_token=access_token or os.getenv("LAKEFS_ACCESS_TOKEN"),
            username=username or os.getenv("LAKEFS_USERNAME") or lakectl_config.username,
            password=password or os.getenv("LAKEFS_PASSWORD") or lakectl_config.password,
            ssl_ca_cert=ssl_ca_cert or os.getenv("LAKEFS_SSL_CA_CERT"),
        )
        # proxy address, not part of the constructor
        configuration.proxy = proxy
        # whether to verify SSL certs, not part of the constructor
        configuration.verify_ssl = verify_ssl

        self.client = LakeFSClient(configuration=configuration)
        self.create_branch_ok = create_branch_ok
        self.source_branch = source_branch

    @cached_property
    def _lakefs_server_version(self):
        with self.wrapped_api_call():
            version_string = self.client.config_api.get_config().version_config.version
            return tuple(int(t) for t in version_string.split("."))

    @classmethod
    @overload
    def _strip_protocol(cls, path: str | os.PathLike[str] | Path) -> str:
        ...

    @classmethod
    @overload
    def _strip_protocol(cls, path: list[str | os.PathLike[str] | Path]) -> list[str]:
        ...

    @classmethod
    def _strip_protocol(cls, path):
        """Copied verbatim from the base class, save for the slash rstrip."""
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        spath = super()._strip_protocol(path)
        if stringify_path(path).endswith("/"):
            return spath + "/"
        return spath

    @property
    def transaction(self):
        """
        A context manager within which file uploads and versioning operations are deferred to a
        queue, and carried out during when exiting the context.

        Requires the file class to implement ``.commit()`` and ``.discard()`` for the normal and exception cases.
        """
        self._transaction: LakeFSTransaction | None
        if self._transaction is None:
            self._transaction = LakeFSTransaction(self)
        return self._transaction

    def start_transaction(self):
        """
        Prepare a lakeFS file system transaction without entering the transaction context yet.
        """
        self._intrans = True
        self._transaction = LakeFSTransaction(self)
        return self.transaction

    @contextmanager
    def wrapped_api_call(
        self, rpath: str | None = None, message: str | None = None, set_cause: bool = True
    ) -> Generator[None, None, None]:
        """
        A context manager to wrap lakeFS API calls, translating any API errors to Python-native OS errors.

        Meant for internal use.

        Parameters
        ----------
        rpath: str | None
            The remote path involved in the requested API call.
        message: str | None
            A custom error message to emit instead of parsing the API error response.
        set_cause: bool
            Whether to include the original lakeFS API error in the resulting traceback.

        Yields
        ------
        None
            An empty generator, to be used as a context manager.

        Raises
        ------
        OSError
            Translated error from the lakeFS API call, if any.
        """
        try:
            yield
        except ApiException as e:
            raise translate_lakefs_error(e, rpath=rpath, message=message, set_cause=set_cause)

    def checksum(self, path: str | os.PathLike[str]) -> str | None:
        """
        Get a remote lakeFS file object's checksum.

        This is usually its MD5 hash, unless another hash function was used on upload.

        Parameters
        ----------
        path: str | os.PathLike[str]
            The remote path to look up the lakeFS checksum for. Must point to a single file object.

        Returns
        -------
        str | None
            The remote file's checksum, or ``None`` if ``path`` points to a directory or does not exist.
        """
        path = stringify_path(path)
        try:
            return self.info(path).get("checksum")
        except FileNotFoundError:
            return None

    def exists(self, path: str | os.PathLike[str], **kwargs: Any) -> bool:
        """
        Check existence of a remote path in a lakeFS repository.

        Input paths can either be files or directories.

        Parameters
        ----------
        path: str | os.PathLike[str]
            The remote path whose existence to check. Must be a fully qualified lakeFS URI.
        **kwargs: Any
            Additional keyword arguments to pass to ``LakeFSClient.objects_api.head_object()``.

        Returns
        -------
        bool
            ``True`` if the requested path exists, ``False`` if it does not.

        Raises
        ------
        PermissionError
            If the user does not have sufficient permissions to query object existence.
        """
        path = stringify_path(path)
        repository, ref, resource = parse(path)

        try:
            self.client.objects_api.head_object(repository, ref, resource, **kwargs)
            return True
        except NotFoundException:
            return False
        except ApiException as e:
            # in case of an error other than "not found", existence cannot be
            # decided, so raise the translated error.
            raise translate_lakefs_error(e)

    def cp_file(
        self, path1: str | os.PathLike[str], path2: str | os.PathLike[str], **kwargs: Any
    ) -> None:
        """
        Copy a single file from one remote location to another in lakeFS.

        Parameters
        ----------
        path1: str | os.PathLike[str]
            The remote file location to be copied.
        path2: str | os.PathLike[str]
            The (remote) target location to which to copy the file.
        **kwargs: Any
            Additional keyword arguments to pass to ``LakeFSClient.objects_api.copy_object()``.

        Raises
        ------
        ValueError
            When attempting to copy objects between repositories.
        """
        path1 = stringify_path(path1)
        path2 = stringify_path(path2)
        if path1 == path2:
            return

        orig_repo, orig_ref, orig_path = parse(path1)
        dest_repo, dest_ref, dest_path = parse(path2)

        if orig_repo != dest_repo:
            raise ValueError(
                "can only copy objects within a repository, but got source "
                f"repository {orig_repo!r} and destination repository {dest_repo!r}"
            )

        with self.wrapped_api_call():
            object_copy_creation = ObjectCopyCreation(src_path=orig_path, src_ref=orig_ref)
            self.client.objects_api.copy_object(
                repository=dest_repo,
                branch=dest_ref,
                dest_path=dest_path,
                object_copy_creation=object_copy_creation,
                **kwargs,
            )

    def get_file(
        self,
        rpath: str | os.PathLike[str],
        lpath: str | os.PathLike[str],
        callback: fsspec.callbacks.Callback = _DEFAULT_CALLBACK,
        outfile: Any = None,
        precheck: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Download a single file from a remote lakeFS server to local storage.

        Parameters
        ----------
        rpath: str | os.PathLike[str]
            The remote path to download to local storage. Must be a fully qualified lakeFS URI, and point to a single file.
        lpath: str | os.PathLike[str]
            The local path on disk to save the downloaded file to.
        callback: fsspec.callbacks.Callback
            An fsspec callback to use during the operation. Can be used to report download progress.
        outfile: Any
            A file-like object to save the downloaded content to. Can be used in place of ``lpath``.
        precheck: bool
            Check if ``lpath`` already exists and compare its checksum with that of ``rpath``, skipping the download if they match.
        **kwargs: Any
            Additional keyword arguments passed to ``AbstractFileSystem.open()``.
        """
        rpath = stringify_path(rpath)
        lpath = stringify_path(lpath)
        lp = Path(lpath)
        if precheck and lp.exists() and lp.is_file():
            local_checksum = md5_checksum(lpath, blocksize=self.blocksize)
            remote_checksum = self.info(rpath).get("checksum")
            if local_checksum == remote_checksum:
                logger.info(
                    f"Skipping download of resource {rpath!r} to local path {lpath!r}: "
                    f"Resource {lpath!r} exists and checksums match."
                )
                return

        with self.wrapped_api_call(rpath=rpath):
            super().get_file(rpath, lpath, callback=callback, outfile=outfile, **kwargs)

    def info(self, path: str | os.PathLike[str], **kwargs: Any) -> dict[str, Any]:
        """
        Query a remote lakeFS object's metadata.

        Parameters
        ----------
        path: str | os.PathLike[str]
            The object for which to obtain metadata. Must be a fully qualified lakeFS URI, can either point to a file or a directory.
        **kwargs: Any
            Additional keyword arguments to pass to either ``LakeFSClient.objects_api.stat_object()``
            (if ``path`` points to a file) or ``LakeFSClient.objects_api.list_objects()`` (if ``path`` points to a directory).

        Returns
        -------
        dict[str, Any]
            A dictionary containing metadata on the object, including its full remote path and object type (file or directory).

        Raises
        ------
        FileNotFoundError
            If the ``path`` refers to a non-file path that does not exist in the repository.
        """
        path = stringify_path(path)
        repository, ref, resource = parse(path)
        # first, try with `stat_object` in case of a file.
        # the condition below checks edge cases of resources that cannot be files.
        if resource and not resource.endswith("/"):
            try:
                # the set of keyword arguments allowed in `list_objects` is a
                # superset of the keyword arguments for `stat_object`.
                # Ensure that only admissible keyword arguments are actually
                # passed to `stat_object`.
                stat_keywords = ["presign", "user_metadata"]
                stat_kwargs = {k: v for k, v in kwargs.items() if k in stat_keywords}

                res = self.client.objects_api.stat_object(
                    repository=repository, ref=ref, path=resource, **stat_kwargs
                )
                return {
                    "checksum": res.checksum,
                    "content-type": res.content_type,
                    "mtime": res.mtime,
                    "name": f"{repository}/{ref}/{res.path}",
                    "size": res.size_bytes,
                    "type": "file",
                }
            except NotFoundException:
                # fall through, retry with `ls` if it's a directory.
                pass
            except ApiException as e:
                raise translate_lakefs_error(e, rpath=path)

        out = self.ls(path, detail=True, recursive=True, **kwargs)
        if not out:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

        return {
            "name": path.rstrip("/"),
            "size": sum(o.get("size") or 0 for o in out),
            "type": "directory",
        }

    def _update_dircache(self, info: list) -> None:
        """Update logic for dircache (optionally recursive) based on lakeFS API response"""
        parents = {self._parent(i["name"]) for i in info}
        for pp in parents:
            # subset of info entries which are direct descendants of `parent`
            dir_info = [i for i in info if self._parent(i["name"].rstrip("/")) == pp]
            if pp not in self.dircache:
                self.dircache[pp] = dir_info
                continue

            # Merge existing dircache entry with updated listing, which contains either:
            # - files not present in the cache yet
            # - a fresh listing (if `refresh=True`)

            cache_entry = self.dircache[pp][:]

            old_names = {e["name"] for e in cache_entry}
            new_names = {e["name"] for e in dir_info}

            to_remove = old_names - new_names
            to_update = old_names.intersection(new_names)

            # Remove all entries no longer present in the current listing
            cache_entry = [e for e in cache_entry if e["name"] not in to_remove]

            # Overwrite existing entries in the cache with its updated values
            for name in to_update:
                old_idx = next(idx for idx, e in enumerate(cache_entry) if e["name"] == name)
                new_entry = next(e for e in info if e["name"] == name)

                cache_entry[old_idx] = new_entry
                dir_info.remove(new_entry)

            # Add the remaining (new) entries to the cache
            cache_entry.extend(dir_info)
            self.dircache[pp] = sorted(cache_entry, key=operator.itemgetter("name"))

    def _ls_from_cache(self, path: str, recursive: bool = False) -> list[dict[str, Any]] | None:
        """Override of ``AbstractFileSystem._ls_from_cache`` with support for recursive listings."""
        if not recursive:
            return super()._ls_from_cache(path)

        result = None
        for key, files in self.dircache.items():
            if not (key.startswith(path) or path == key + "/"):
                continue
            if result is None:
                result = []
            result.extend(files)
        if not result:
            return result
        return sorted(result, key=operator.itemgetter("name"))

    @overload
    def ls(
        self,
        path: str | os.PathLike[str],
        detail: Literal[True] = ...,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        ...

    @overload
    def ls(
        self,
        path: str | os.PathLike[str],
        detail: Literal[False],
        **kwargs: Any,
    ) -> list[str]:
        ...

    @overload
    def ls(
        self,
        path: str | os.PathLike[str],
        detail: bool = True,
        **kwargs: Any,
    ) -> list[str] | list[dict[str, Any]]:
        ...

    def ls(
        self,
        path: str | os.PathLike[str],
        detail: bool = True,
        **kwargs: Any,
    ) -> list[str] | list[dict[str, Any]]:
        """
        List all available objects under a given path in lakeFS.

        Parameters
        ----------
        path: str | os.PathLike[str]
            The path under which to list objects. Must be a fully qualified lakeFS URI.
            Can also point to a file, in which case the file's metadata will be returned.
        detail: bool
            Whether to obtain all metadata on the requested objects or just their names.
        **kwargs: Any
            Additional keyword arguments to pass to ``LakeFSClient.objects_api.list_objects()``.

            In particular:
                `refresh: bool`: whether to skip the directory listing cache,
                `recursive: bool`: whether to list subdirectory contents recursively

        Returns
        -------
        list[str] | list[dict[str, Any]]
            A list of all objects' metadata under the given remote path if ``detail=True``, or alternatively only their names if ``detail=False``.
        """

        def _api_path_type_to_info(path_type: str) -> Literal["file", "directory"]:
            """Convert ``list_objects()`` API response field ``path_type`` to ``info.type``."""
            if path_type == "object":
                return "file"
            elif path_type == "common_prefix":
                return "directory"
            else:
                raise ValueError(f"unexpected path type {path_type!r}")

        path = cast(str, stringify_path(path))
        repository, ref, prefix = parse(path)

        recursive = kwargs.pop("recursive", False)

        # Try lookup in dircache unless explicitly disabled by `refresh=True` kwarg
        use_dircache = not kwargs.pop("refresh", False)

        if use_dircache:
            cache_entry: list[Any] | None = None
            try:
                cache_entry = self._ls_from_cache(path, recursive=recursive)
            except FileNotFoundError:
                # we patch files missing from an ls call in the cache entry below,
                # so this should not be an error.
                pass

            if cache_entry is not None:
                if not detail:
                    return [e["name"] for e in cache_entry]
                return cache_entry[:]

        kwargs["prefix"] = prefix

        info = []
        # stat infos are either the path only (`detail=False`) or a dict full of metadata
        with self.wrapped_api_call(rpath=path):
            delimiter = "" if recursive else "/"
            objects = depaginate(
                self.client.objects_api.list_objects,
                repository,
                ref,
                delimiter=delimiter,
                **kwargs,
            )
            for obj in cast(Iterable[ObjectStats], objects):
                info.append(
                    {
                        "checksum": obj.checksum,
                        "content-type": obj.content_type,
                        "mtime": obj.mtime,
                        "name": f"{repository}/{ref}/{obj.path}",
                        "size": obj.size_bytes,
                        "type": _api_path_type_to_info(obj.path_type),
                    }
                )

        # Retry the API call with appended slash if the current result
        # is just a single directory entry only (not its contents).
        # This is useful to allow `ls("repo/branch/dir")` calls without
        # a trailing slash.
        if len(info) == 1 and info[0]["type"] == "directory":
            return self.ls(
                path + "/",
                detail=detail,
                **kwargs | {"refresh": not use_dircache, "recursive": recursive},
            )

        if recursive:
            # To make recursive ls behave identical to the non-recursive case,
            # add back virtual `directory` entries, which are only returned by
            # the lakeFS API when querying non-recursively.
            here = self._strip_protocol(path).rstrip("/")
            subdirs = {parent for o in info if (parent := self._parent(o["name"])) != here}
            for subdir in subdirs:
                info.append(
                    {
                        "type": "directory",
                        "name": subdir + "/",
                        "size": 0,
                    }
                )

        if info:
            self._update_dircache(info[:])

        if not detail:
            info = [o["name"] for o in info]

        return info

    def _open(
        self,
        path: str,
        mode: Literal["rb", "wb"] = "rb",
        block_size: int | None = None,
        autocommit: bool = True,
        cache_options: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> LakeFSFile:
        """
        Dispatch a lakeFS file (local buffer on disk) for the given remote path for up- or downloads depending on ``mode``.

        Internal only, called by ``AbstractFileSystem.open()``.

        Parameters
        ----------
        path: str
            The remote path for which to open a local ``LakeFSFile``. Must be a fully qualified lakeFS URI.
        mode: Literal["rb", "wb"]
            The file mode indicating its purpose. Use ``rb`` for downloads from lakeFS, ``wb`` for uploads to lakeFS.
        block_size: int | None
            The file block size to read at a time. If not set, falls back to fsspec's default blocksize of 5 MB.
        autocommit: bool
            Whether to write the file buffer automatically on file closing in write mode.
        cache_options: dict[str, str] | None
            Additional caching options to pass to the ``AbstractBufferedFile`` superclass.
        **kwargs: Any
            Additional keyword arguments to pass to ``LakeFSClient.objects_api.get_object()`` on download (``mode = 'rb'``),
            or ``LakeFSClient.objects_api.put_object()`` on upload (``mode='wb'``).

        Returns
        -------
        LakeFSFile
            A local file-like object ready to hold data to be received from / sent to a lakeFS server.

        Raises
        ------
        NotImplementedError
            If ``mode`` is not supported.
        """
        if mode not in {"rb", "wb"}:
            raise NotImplementedError(f"unsupported mode {mode!r}")

        return LakeFSFile(
            self,
            path=path,
            mode=mode,
            block_size=block_size or self.blocksize,
            autocommit=autocommit,
            cache_options=cache_options,
            **kwargs,
        )

    def put_file_to_blockstore(
        self,
        lpath: str | os.PathLike[str],
        rpath: str | os.PathLike[str],
        callback: fsspec.callbacks.Callback = _DEFAULT_CALLBACK,
        presign: bool = False,
        storage_options: dict[str, Any] | None = None,
    ) -> None:
        """
        Upload a file to lakeFS by directly putting it into its underlying block storage, thereby reducing the request load
        on the lakeFS server.

        Requires the corresponding fsspec implementation for the block storage type used by your lakeFS server deployment.

        Supported block storage types are S3 (needs ``s3fs``), GCS (needs ``gcsfs``), and Azure Blob Storage (needs ``adlfs``).

        Note that depending on the block store type, additional configuration like credentials may need to be configured when ``presign=False``.

        Parameters
        ----------
        lpath: str | os.PathLike[str]
            The local path to upload to the lakeFS block storage.
        rpath: str | os.PathLike[str]
            The remote target path to upload the local file to. Must be a fully qualified lakeFS URI.
        callback: fsspec.callbacks.Callback
            An fsspec callback to use during the operation. Can be used to report download progress.
        presign: bool
            Whether to use pre-signed URLs to upload the object via HTTP(S) using ``urllib.request``.
        storage_options: dict[str, Any] | None
            Additional file system configuration options to pass to the block storage file system.

        Raises
        ------
        ValueError
            If the blockstore type returned by the lakeFS API is not supported by fsspec.
        """
        rpath = stringify_path(rpath)
        lpath = stringify_path(lpath)
        repository, branch, resource = parse(rpath)

        staging_location = self.client.staging_api.get_physical_address(
            repository, branch, resource, presign=presign
        )

        if presign:
            remote_url = staging_location.presigned_url
            content_type, _ = mimetypes.guess_type(lpath)
            if content_type is None:
                content_type = "application/octet-stream"
            with open(lpath, "rb") as f:
                headers = {
                    "Content-Type": content_type,
                }
                request = urllib.request.Request(
                    url=remote_url, data=f, headers=headers, method="PUT"
                )
                try:
                    if not remote_url.lower().startswith("http"):
                        raise ValueError("Wrong protocol for remote connection")
                    else:
                        logger.debug(f"Begin upload of {lpath}")
                        with urllib.request.urlopen(request):  # nosec [B310:blacklist] # We catch faulty protocols above.
                            logger.debug(f"Successfully uploaded {lpath}")
                except urllib.error.HTTPError as e:
                    raise translate_lakefs_error(e, rpath=rpath)
        else:
            blockstore_type = self.client.config_api.get_config().storage_config.blockstore_type
            # lakeFS blockstore name is "azure", but Azure's fsspec registry entry is "az".
            if blockstore_type == "azure":
                blockstore_type = "az"

            if blockstore_type not in ["s3", "gs", "az"]:
                raise ValueError(
                    f"Blockstore writes are not implemented for blockstore type {blockstore_type!r}"
                )

            remote_url = staging_location.physical_address
            remote = filesystem(blockstore_type, **(storage_options or {}))
            remote.put_file(lpath, remote_url, callback=callback)

        staging_metadata = StagingMetadata(
            staging=staging_location,
            checksum=md5_checksum(lpath, blocksize=self.blocksize),
            size_bytes=os.path.getsize(lpath),
        )
        self.client.staging_api.link_physical_address(
            repository, branch, resource, staging_metadata
        )

    def put_file(
        self,
        lpath: str | os.PathLike[str],
        rpath: str | os.PathLike[str],
        callback: fsspec.callbacks.Callback = _DEFAULT_CALLBACK,
        precheck: bool = True,
        use_blockstore: bool = False,
        presign: bool = False,
        storage_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Upload a local file to a remote location on a lakeFS server.

        Note that depending on the block store type, additional configuration like credentials may need to be configured when ``use_blockstore=True`` and ``presign=False``.

        Parameters
        ----------
        lpath: str | os.PathLike[str]
            The local path on disk to upload to the lakeFS server.
        rpath: str | os.PathLike[str]
            The remote target path to upload the local file to. Must be a fully qualified lakeFS URI.
        callback: fsspec.callbacks.Callback
            An fsspec callback to use during the operation. Can be used to report download progress.
        precheck: bool
            Check if ``lpath`` already exists and compare its checksum with that of ``rpath``, skipping the download if they match.
        use_blockstore: bool
            Optionally upload the file directly to the underlying block storage, thereby bypassing the lakeFS server and saving a
            file transfer. Preferable for uploads of large files.
        presign: bool
            Whether to use pre-signed URLs to upload the object if ``use_blockstore=True``.
        storage_options: dict[str, Any] | None
            Additional file system configuration options to pass to the block storage file system if ``use_blockstore=True``.
        **kwargs: Any
            Additional keyword arguments to pass to ``AbstractFileSystem.open()``.
        """
        lpath = stringify_path(lpath)
        rpath = stringify_path(rpath)
        if precheck and Path(lpath).is_file():
            remote_checksum = self.checksum(rpath)
            local_checksum = md5_checksum(lpath, blocksize=self.blocksize)
            if local_checksum == remote_checksum:
                logger.info(
                    f"Skipping upload of resource {lpath!r} to remote path {rpath!r}: "
                    f"Resource {rpath!r} exists and checksums match."
                )
                return

        if use_blockstore:
            self.put_file_to_blockstore(
                lpath,
                rpath,
                presign=presign,
                callback=callback,
                storage_options=storage_options,
            )
        else:
            with self.wrapped_api_call(rpath=rpath):
                super().put_file(lpath, rpath, callback=callback, **kwargs)

    def rm_file(self, path: str | os.PathLike[str]) -> None:
        """
        Stage a remote file for removal on a lakeFS server.

        The file will not actually be removed from the requested branch until a commit is created.

        Parameters
        ----------
        path: str | os.PathLike[str]
            The remote file to delete. Must be a fully qualified lakeFS URI.
        """
        path = stringify_path(path)
        repository, branch, resource = parse(path)

        with self.wrapped_api_call(rpath=path):
            self.client.objects_api.delete_object(
                repository=repository, branch=branch, path=resource
            )
            # Directory listing cache for the containing folder must be invalidated
            self.dircache.pop(self._parent(path), None)

    def touch(self, path: str | os.PathLike[str], truncate: bool = True, **kwargs: Any) -> None:
        """
        Create an empty file or update an existing file on a lakeFS server.

        Parameters
        ----------
        path: str | os.PathLike[str]
            The file path to create or update. Must be a fully qualified lakeFS URI.
        truncate: bool
            Whether to set the file size to 0 (zero) bytes, even if the path already exists.
        **kwargs: Any
            Additional keyword arguments to pass to ``LakeFSFile.open()``.

        Raises
        ------
        NotImplementedError
            If the targeted lakeFS server version does not support `touch()` operations.
        """

        # empty buffer upload errors were fixed in https://github.com/treeverse/lakeFS/issues/7130,
        # which was first released in lakeFS v1.3.1.
        if self._lakefs_server_version < (1, 3, 1):
            version_string = ".".join(str(v) for v in self._lakefs_server_version)
            raise NotImplementedError(
                "LakeFSFileSystem.touch() is not supported for your lakeFS server version. "
                f"minimum required version: '1.3.1', actual version: {version_string!r}"
            )

        super().touch(path=path, truncate=truncate, **kwargs)


class LakeFSFile(AbstractBufferedFile):
    """
    lakeFS file implementation.

    Notes
    -----
    Creates a local buffer on disk for the given remote path for up- or downloads depending on ``mode``.

    Read operations are buffered, write operations are unbuffered. This means that local files to be uploaded will be loaded entirely into memory.

    Parameters
    ----------
    fs: LakeFSFileSystem
        The lakeFS file system associated to this file.
    path: str | os.PathLike[str]
        The remote path to either up- or download depending on ``mode``. Must be a fully qualified lakeFS URI.
    mode: Literal["rb", "wb"]
        The file mode indicating its purpose. Use ``rb`` for downloads from lakeFS, ``wb`` for uploads to lakeFS.
    block_size: int | str
        The file block size to read at a time. If not set, falls back to fsspec's default blocksize of 5 MB.
    autocommit: bool
        Whether to write the file buffer automatically to lakeFS on file closing in write mode.
    cache_type: str
        Cache policy in read mode (any of ``readahead``, ``none``, ``mmap``, ``bytes``). See ``AbstractBufferedFile`` for details.
    cache_options: dict[str, Any] | None
        Additional options passed to the constructor for the cache specified by ``cache_type``.
    size: int | None
        If given and ``mode='rb'``, this will be used as the file size (in bytes) instead of determining it from the remote file.
    **kwargs: Any
        Additional keyword arguments to pass to ``LakeFSClient.objects_api.get_object()`` on download (``mode='rb'``),
        or ``LakeFSClient.objects_api.put_object()`` on upload (``mode='wb'``).
    """

    def __init__(
        self,
        fs: LakeFSFileSystem,
        path: str | os.PathLike[str],
        mode: Literal["rb", "wb"] = "rb",
        block_size: int | str = "default",
        autocommit: bool = True,
        cache_type: str = "readahead",
        cache_options: dict[str, Any] | None = None,
        size: int | None = None,
        **kwargs: Any,
    ):
        path = stringify_path(path)
        super().__init__(
            fs,
            path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_type=cache_type,
            cache_options=cache_options,
            size=size,
            **kwargs,
        )

        self.buffer: io.BytesIO
        if mode == "wb" and self.fs.create_branch_ok:
            repository, branch, resource = parse(path)
            create_branch(self.fs.client, repository, branch, self.fs.source_branch)

    def __del__(self):
        """Custom deleter, only here to unset the base class behavior."""
        pass

    def _upload_chunk(self, final: bool = False) -> bool:
        """
        Commit the file on final chunk via single-shot upload, no-op otherwise.

        Parameters
        ----------
        final: bool
            Proceed with uploading the file if ``self.autocommit=True``.

        Returns
        -------
        bool
            If the file buffer needs more data to be written before initiating the upload.
        """
        if final and self.autocommit:
            self.commit()
        return not final

    def commit(self):
        """
        Upload the file to lakeFS in single-shot mode.

        Results in an unbuffered upload, and a memory allocation in the magnitude of the file size on the caller's host machine.
        """
        repository, branch, resource = parse(self.path)

        with self.fs.wrapped_api_call(rpath=self.path):
            # empty buffer is equivalent to a touch()
            self.buffer.seek(0)
            self.fs.client.objects_api.upload_object(
                repository=repository,
                branch=branch,
                path=resource,
                content=self.buffer.read(),
                **self.kwargs,
            )

        self.buffer = io.BytesIO()

    def discard(self):
        """Discard the file's current buffer."""
        self.buffer = io.BytesIO()  # discards the data, but in a type-safe way.

    def flush(self, force: bool = False) -> None:
        """
        Write buffered data to backend store.

        Writes the current buffer, if it is larger than the block-size, or if
        the file is being closed.

        In contrast to the abstract class, this implementation does NOT unload the buffer
        if it is larger than the block size, because the lakeFS server does not support
        multipart uploads.

        Parameters
        ----------
        force: bool
            When closing, write the last block even if it is smaller than
            blocks are allowed to be. Disallows further writing to this file.

        Raises
        ------
        ValueError
            If the file is closed, or has already been forcibly flushed and ``force=True``.
        """

        if self.closed:
            raise ValueError("Flush on closed file")
        self.forced: bool
        if force and self.forced:
            raise ValueError("Force flush cannot be called more than once")
        if force:
            self.forced = True

        if self.mode != "wb":
            # no-op to flush on read-mode
            return

        if not force and self.buffer.tell() < self.blocksize:
            # Defer write on small block
            return

        self.offset: int
        if self.offset is None:
            # Initialize an upload
            self.offset = 0

        if self._upload_chunk(final=force) is not False:
            self.offset += self.buffer.seek(0, 2)

    def _fetch_range(self, start: int, end: int) -> bytes:
        """
        Fetch a byte range of the ``LakeFSFile``'s target remote path.

        The byte range is right-exclusive, meaning that the amount of transferred bytes equals ``end - start``.

        Parameters
        ----------
        start: int
            Start of the byte range, inclusive.
        end: int
            End of the byte range, exclusive. Must be greater than ``start``.

        Returns
        -------
        bytes
            A byte array holding the downloaded data from lakeFS.
        """
        repository, ref, resource = parse(self.path)
        with self.fs.wrapped_api_call(rpath=self.path):
            return self.fs.client.objects_api.get_object(
                repository, ref, resource, range=f"bytes={start}-{end - 1}", **self.kwargs
            )



import pytest

from lakefs_spec import LakeFSFileSystem


def test_info_on_directory(fs: LakeFSFileSystem, repository: str) -> None:
    """
    Check that an `fs.info` call on a trailing-slash resource yields a directory.
    """
    resource = f"{repository}/main/"
    res = fs.info(resource)

    assert res["type"] == "directory"


def test_info_on_nonexistent_directory(fs: LakeFSFileSystem, repository: str) -> None:
    """
    Check that a nonexistent directory raises a FileNotFoundError in `fs.info`.
    """
    resource = f"{repository}/main/blabla/"

    with pytest.raises(FileNotFoundError):
        fs.info(resource)


def test_info_on_directory_no_trailing_slash(fs: LakeFSFileSystem, repository: str) -> None:
    """
    Regression test to check that calling `fs.info()` on existing non-slash-terminated
    directories yields the same results as if terminated with a slash.
    """
    resource = f"{repository}/main/data/"
    res = fs.info(resource)

    assert res["type"] == "directory"

    non_slash_resource = resource.rstrip("/")
    assert fs.info(non_slash_resource) == res



"""
Functionality for working with ``lakectl`` configuration files useable for authentication in the lakeFS file system.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, NamedTuple

import yaml


class LakectlConfig(NamedTuple):
    """
    Holds configuration values necessary for authentication with a lakeFS server from Python.
    """

    host: str | None = None
    """URL of the lakeFS host, http(s) prefix is optional."""
    username: str | None = None
    """The access key ID to use in authentication with lakeFS."""
    password: str | None = None
    """The secret access key to use in authentication with lakeFS."""

    @classmethod
    def read(cls, path: str | Path) -> "LakectlConfig":
        """
        Read in a lakectl YAML configuration file and parse out relevant authentication parameters.

        Parameters
        ----------
        path: str | Path
            Path to the YAML configuration file.

        Returns
        -------
        LakectlConfig
            The immutable loaded configuration. Missing values are filled with ``None`` placeholders.

        Raises
        ------
        FileNotFoundError
            If the configuration file does not exist.
        """
        if not Path(path).exists():
            raise FileNotFoundError(path)

        with open(path, "r") as f:
            obj: dict[str, Any] = yaml.safe_load(f)

        # config struct schema (Golang backend code):
        # https://github.com/treeverse/lakeFS/blob/master/cmd/lakectl/cmd/root.go
        creds: dict[str, str] = obj.get("credentials", {})
        server: dict[str, str] = obj.get("server", {})
        username = creds.get("access_key_id")
        password = creds.get("secret_access_key")
        host = server.get("endpoint_url")
        return cls(host=host, username=username, password=password)



from typing import Any

import pytest

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory, with_counter


def test_copy(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
    temporary_branch_context: Any,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    fs.put(lpath, rpath)
    assert fs.exists(rpath)

    with temporary_branch_context("new-copy-test") as b:
        new_rpath = f"{repository}/{b}/{random_file.name}"
        fs.cp_file(rpath, new_rpath)
        assert fs.exists(new_rpath)


def test_copy_edge_cases(
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    fs.client, counter = with_counter(fs.client)

    path = f"{repository}/main/lakes.parquet"

    fs.cp_file(path, path)

    assert counter.count("objects_api.copy_object") == 0

    with pytest.raises(ValueError, match="can only copy objects within a repository.*"):
        fs.cp_file(path, "my-repo/main/lakes.parquet")



import collections
import random
import string
import uuid
from inspect import ismethod
from pathlib import Path
from typing import Optional, Union

from lakefs_sdk.client import LakeFSClient

import lakefs_spec.client_helpers as client_helpers
from lakefs_spec import LakeFSFileSystem


class APICounter:
    def __init__(self):
        self._counts: dict[str, int] = collections.defaultdict(int)

    def clear(self):
        self._counts.clear()

    def count(self, name: str) -> int:
        return self._counts[name]

    def counts(self):
        """Gives an iterator over the API counts."""
        return self._counts.values()

    def named_counts(self):
        """Gives an iterator over the API names and counts."""
        return self._counts.items()

    def increment(self, name: str) -> None:
        self._counts[name] += 1


def with_counter(client: LakeFSClient) -> tuple[LakeFSClient, APICounter]:
    """Instruments a lakeFS API client with an API counter."""
    counter = APICounter()

    def patch(fn, name):
        """Patches an API instance method ``fn`` on an API ``apiname``."""

        def wrapped_fn(*args, **kwargs):
            counter.increment(name)
            return fn(*args, **kwargs)

        return wrapped_fn

    for api_name, api in client.__dict__.items():
        if api_name == "_api":
            continue

        for ep_name in filter(
            lambda op: not op.startswith("_")
            and not op.endswith("with_http_info")
            and ismethod(getattr(api, op)),
            dir(api),
        ):
            endpoint = getattr(api, ep_name)
            setattr(api, ep_name, patch(endpoint, f"{api_name}.{ep_name}"))

    return client, counter


class RandomFileFactory:
    def __init__(self, path: Union[str, Path]):
        path = Path(path)
        if not path.is_dir():
            raise ValueError(f"input path needs to be a directory, got {path}")

        self.path = path

    def list(self) -> list[Path]:
        return list(self.path.iterdir())

    def make(self, fname: Optional[str] = None, size: int = 2**10) -> Path:
        """
        Generate a random file named ``fname`` with a random string of size
        ``size`` (in bytes) as content.
        """
        if fname is None:
            fname = "test-" + str(uuid.uuid4()) + ".txt"
        random_file = self.path / fname
        random_str = "".join(random.choices(string.ascii_letters + string.digits, k=size))
        random_file.write_text(random_str, encoding="utf-8")
        return random_file


def commit_random_file_on_branch(
    random_file_factory: RandomFileFactory, fs: LakeFSFileSystem, repository: str, temp_branch: str
) -> None:
    random_file = random_file_factory.make()
    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    fs.put(lpath, rpath, precheck=False)

    commit = client_helpers.commit(
        client=fs.client,
        repository=repository,
        branch=temp_branch,
        message=f"Add file {random_file.name!r}",
    )


def list_branch_names(fs: LakeFSFileSystem, repository: str) -> list[str]:
    return [branch.id for branch in client_helpers.list_branches(fs.client, repository=repository)]



from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory


def test_exists(fs: LakeFSFileSystem, repository: str) -> None:
    """Test `fs.exists` on an existing file. Requires a populated repository."""

    example_file = f"{repository}/main/README.md"
    assert fs.exists(example_file)

    nonexistent_file = f"{repository}/main/nonexistent.parquet"
    assert not fs.exists(nonexistent_file)


def test_exists_on_staged_file(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: str,
    temp_branch: str,
) -> None:
    random_file = random_file_factory.make()

    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    # upload, verify existence.
    fs.put(lpath, rpath)
    assert fs.exists(rpath)



import collections
from unittest.mock import MagicMock, patch

import pytest
from lakefs_sdk.models.staging_location import StagingLocation

from lakefs_spec import LakeFSFileSystem
from tests.util import RandomFileFactory

# Mock a lakeFS config return object.
# we only query the block store type for now, we can add more attributes later if needed.
# we define these mock types because it is way easier than initializing the proper
# pydantic models with mock data.
StorageConfig = collections.namedtuple("StorageConfig", ["blockstore_type"])
ConfigMock = collections.namedtuple("ConfigMock", ["storage_config"])


def test_local_blockstore_type(
    random_file_factory: RandomFileFactory, repository: str, temp_branch: str, fs: LakeFSFileSystem
) -> None:
    random_file = random_file_factory.make()
    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    with patch(
        "lakefs_sdk.api.ConfigApi.get_config", new_callable=MagicMock
    ) as mock_get_client_blockstore_type:
        mock_get_client_blockstore_type.return_value = ConfigMock(
            storage_config=StorageConfig("local")
        )
        with pytest.raises(ValueError, match="not implemented for blockstore type 'local'"):
            fs.put_file_to_blockstore(lpath, rpath)


def test_presigned_url(
    mock_urlopen: MagicMock,
    random_file_factory: RandomFileFactory,
    repository: str,
    temp_branch: str,
    fs: LakeFSFileSystem,
) -> None:
    random_file = random_file_factory.make()
    lpath = str(random_file)
    rpath = f"{repository}/{temp_branch}/{random_file.name}"

    with patch(
        "lakefs_sdk.api.ConfigApi.get_config", new_callable=MagicMock
    ) as mock_get_client_blockstore_type:
        mock_get_client_blockstore_type.return_value = ConfigMock(
            storage_config=StorageConfig("s3")
        )
        get_physical_address_mock = fs.client.staging_api.get_physical_address = MagicMock(
            return_value=StagingLocation(presigned_url="http://mock.address", token="123")
        )
        link_physical_address_mock = fs.client.staging_api.link_physical_address = MagicMock()

        fs.put_file_to_blockstore(lpath, rpath, presign=True)
    mock_urlopen.assert_called_once()
    get_physical_address_mock.assert_called_once()
    link_physical_address_mock.assert_called_once()



"""
Useful utilities for handling lakeFS URIs and results of lakeFS API calls.
"""
from __future__ import annotations

import hashlib
import os
import re
from typing import Any, Callable, Generator, Protocol

from lakefs_sdk import Pagination
from lakefs_sdk import __version__ as __lakefs_sdk_version__

lakefs_sdk_version = tuple(int(v) for v in __lakefs_sdk_version__.split("."))
del __lakefs_sdk_version__


class PaginatedApiResponse(Protocol):
    pagination: Pagination
    results: list


def depaginate(
    api: Callable[..., PaginatedApiResponse], *args: Any, **kwargs: Any
) -> Generator[Any, None, None]:
    """
    Unwrap the responses from a paginated lakeFS API method into a generator.

    Parameters
    ----------
    api: Callable[..., PaginatedApiResponse]
        The lakeFS client API to call. Must return a paginated response with the ``pagination`` and ``results`` fields set.
    *args: Any
        Positional arguments to pass to the API call.
    **kwargs: Any
        Keyword arguments to pass to the API call.

    Yields
    ------
    Any
        The obtained API result objects.
    """
    while True:
        resp = api(*args, **kwargs)
        yield from resp.results
        if not resp.pagination.has_more:
            break
        kwargs["after"] = resp.pagination.next_offset


def md5_checksum(lpath: str | os.PathLike[str], blocksize: int = 2**22) -> str:
    """
    Calculate a local file's MD5 hash.

    Parameters
    ----------
    lpath: str | os.PathLike[str]
        The local path whose MD5 hash to calculate. Must be a file.
    blocksize: int
        Block size (in bytes) to use while reading in the file.

    Returns
    -------
    str
        The file's MD5 hash value, as a string.
    """
    with open(lpath, "rb") as f:
        file_hash = hashlib.md5(usedforsecurity=False)
        chunk = f.read(blocksize)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(blocksize)
    return file_hash.hexdigest()


def parse(path: str) -> tuple[str, str, str]:
    """
    Parses a lakeFS URI in the form ``lakefs://<repo>/<ref>/<resource>``.

    Parameters
    ----------
    path: str
        String path, needs to conform to the lakeFS URI format described above.
        The ``<resource>`` part can be the empty string; the leading ``lakefs://`` scheme may be omitted.

    Returns
    -------
    tuple[str, str, str]
        A 3-tuple of repository name, reference, and resource name.

    Raises
    ------
    ValueError
        If the path does not conform to the lakeFS URI format.
    """

    # First regex reflects the lakeFS repository naming rules:
    # only lowercase letters, digits and dash, no leading dash, minimum 3, maximum 63 characters
    # https://docs.lakefs.io/understand/model.html#repository
    # Second regex is the branch: Only letters, digits, underscores and dash, no leading dash.
    path_regex = re.compile(r"(?:lakefs://)?([a-z0-9][a-z0-9\-]{2,62})/(\w[\w\-]*)/(.*)")
    results = path_regex.fullmatch(path)
    if results is None:
        raise ValueError(
            f"expected path with structure lakefs://<repo>/<ref>/<resource>, got {path!r}"
        )

    repo, ref, resource = results.groups()
    return repo, ref, resource



from pathlib import Path

from lakefs_spec import LakeFSFileSystem

REPO, BRANCH = "repo", "main"

# Prepare example local data
local_path = Path("demo.txt")
local_path.write_text("Hello, lakeFS!")

# Upload the local file to the repo and commit
fs = LakeFSFileSystem()  # will auto-discover credentials from ~/.lakectl.yaml
repo_path = f"{REPO}/{BRANCH}/{local_path.name}"

with fs.transaction as tx:
    fs.put(str(local_path), repo_path)
    tx.commit(REPO, BRANCH, "Add demo data")

# Read back the file contents
f = fs.open(repo_path, "rt")
print(f.readline())  # prints "Hello, lakeFS!"

# Compare the sizes of local file and repo
file_info = fs.info(repo_path)
print(
    f"{local_path.name}: local size: {file_info['size']}, remote size: {local_path.stat().st_size}"
)

# Get information about all files in the repo root
print(fs.ls(f"{REPO}/{BRANCH}/"))

# Delete uploaded file from the repository (and commit)
with fs.transaction as tx:
    fs.rm(repo_path)
    tx.commit(REPO, BRANCH, "Delete demo data")
local_path.unlink()



import pandas as pd

from lakefs_spec import LakeFSFileSystem

fs = LakeFSFileSystem()

with fs.transaction as tx:
    tx.create_branch("quickstart", "german-lakes", "main")

    lakes = pd.read_parquet("lakefs://quickstart/main/lakes.parquet")
    german_lakes = lakes.query('Country == "Germany"')
    german_lakes.to_csv("lakefs://quickstart/german-lakes/german_lakes.csv")

    tx.commit("quickstart", "german-lakes", "Add German lakes")
